"""iOS .IPA asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector("v3.asset.file.ios.ipa")
class IOSIpa(asset.Asset):
    """IOS .IPA target asset."""

    def __init__(
        self,
        content: Optional[bytes] = None,
        path: Optional[str] = None,
        content_url: Optional[str] = None,
    ):
        self.content = content
        self.path = path
        self.content_url = content_url

    def __str__(self) -> str:
        str_representation = "iOS"
        if self.path is not None:
            str_representation = f"{str_representation}:{self.path}"
        if self.content_url is not None:
            str_representation = f"{str_representation}:{self.content_url}"

        return str_representation

    @property
    def proto_field(self) -> str:
        return "ios_ipa"



"""Ostorlab lite local runtime modules. this module takes care of launching a scan in the
local machine but with little setup, like no MQ service, no scan persistence and no defaults
agents are started."""
from .runtime import LiteLocalRuntime



"""Asset of type agent for meta-scanning."""
import logging
from typing import Optional

import click

from ostorlab.assets import agent as agent_asset
from ostorlab.cli.scan.run import run

logger = logging.getLogger(__name__)


@run.run.command()
@click.option("--key", help="Agent key.", required=True)
@click.option("--version", help="Agent version.", required=False)
@click.option("--git-location", help="Agent source Git repo.", required=False)
@click.option("--docker-location", help="Agent store docker URL.", required=False)
@click.option(
    "--yaml-file-location",
    help="Agent YAML definition path in the Git repo.",
    required=False,
)
@click.pass_context
def agent(
    ctx: click.core.Context,
    key: str,
    version: Optional[str] = None,
    git_location: Optional[str] = None,
    docker_location: Optional[str] = None,
    yaml_file_location: Optional[str] = None,
) -> None:
    """Run scan for agent."""
    runtime = ctx.obj["runtime"]
    asset = agent_asset.Agent(
        key=key,
        version=version,
        git_location=git_location,
        docker_location=docker_location,
        yaml_file_location=yaml_file_location,
    )
    logger.debug("scanning asset %s", asset)
    runtime.scan(
        title=ctx.obj["title"],
        agent_group_definition=ctx.obj["agent_group_definition"],
        assets=[asset],
    )



"""KB json files test"""
import glob
import json
import pathlib

import pytest


def testJsonFiles_allFilesAreValid_testPasses() -> None:
    json_files = glob.glob("**/*.json", recursive=True)

    for json_file in json_files:
        with pathlib.Path(json_file).open(encoding="utf-8") as file:
            try:
                json_data = json.load(file)
            except ValueError as e:
                pytest.fail(f"Failed to load JSON file '{json_file}': {str(e)}")

            # Check if the JSON data is a dictionary
            assert isinstance(json_data, dict), "JSON data must be a dictionary."

            # Check if the required keys are present
            required_keys = ["risk_rating", "short_description", "references", "title"]
            for key in required_keys:
                assert (
                    key in json_data
                ), f"Required key '{key}' is missing in JSON data."

            # Check the data types and formats of the keys
            assert isinstance(
                json_data["risk_rating"], str
            ), "risk_rating must be a string."
            assert isinstance(
                json_data["short_description"], str
            ), "short_description must be a string."
            assert isinstance(
                json_data["references"], dict
            ), "references must be a dictionary."
            assert isinstance(json_data["title"], str), "title must be a string."

            # Check the format of the references
            references = json_data["references"]
            assert isinstance(references, dict), "references must be a dictionary."



"""Common logic for Mobile Assets of type .APK package file AAB package file and IPA .
This module takes care of preparing the application file and calling the create mobile scan API.
"""
import io
import click
import itertools

from typing import List

from ostorlab.cli.ci_scan.run import run
from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis.runners import runner as base_runner
from ostorlab.apis import scan_create as scan_create_api
from ostorlab.apis import test_credentials_create as test_credentials_create_api


def _prepare_test_credentials(
    ctx: click.core.Context,
) -> List[test_credentials_create_api.TestCredential]:
    test_credentials_login = ctx.obj["test_credentials"]["test_credentials_login"]
    test_credentials_password = ctx.obj["test_credentials"]["test_credentials_password"]
    test_credentials_url = ctx.obj["test_credentials"]["test_credentials_url"]
    test_credentials_role = ctx.obj["test_credentials"]["test_credentials_role"]
    test_credentials_name = ctx.obj["test_credentials"]["test_credentials_name"]
    test_credentials_value = ctx.obj["test_credentials"]["test_credentials_value"]

    credentials = []
    for login, password, role, url in itertools.zip_longest(
        test_credentials_login,
        test_credentials_password,
        test_credentials_role,
        test_credentials_url,
    ):
        credentials.append(
            test_credentials_create_api.TestCredentialLogin(
                login=login, password=password, role=role, url=url
            )
        )

    if test_credentials_name:
        credentials.append(
            test_credentials_create_api.TestCredentialCustom(
                values=dict(
                    itertools.zip_longest(test_credentials_name, test_credentials_value)
                )
            )
        )

    return credentials


def run_mobile_scan(
    ctx: click.core.Context,
    file: io.FileIO,
    asset_type: scan_create_api.MobileAssetType,
) -> None:
    """Create scan for mobile application package file."""
    ci_logger = ctx.obj["ci_logger"]
    if ctx.obj.get("api_key"):
        scan_profile = ctx.obj["scan_profile"]
        title = ctx.obj["title"]
        break_on_risk_rating = ctx.obj["break_on_risk_rating"]
        max_wait_minutes = ctx.obj["max_wait_minutes"]
        sboms = ctx.obj["sboms"]
        runner = authenticated_runner.AuthenticatedAPIRunner(
            api_key=ctx.obj.get("api_key")
        )
        try:
            test_credentials = _prepare_test_credentials(ctx)
            if test_credentials:
                credential_ids = _create_test_credentials(test_credentials, runner)
                ci_logger.info(
                    f'Created test credentials {", ".join(str(t) for t in test_credentials)} successfully'
                )
            else:
                credential_ids = []

            ci_logger.info(
                f"creating scan `{title}` with profile `{scan_profile}` for `{asset_type}`"
            )
            scan_id = _create_scan(
                title,
                scan_profile,
                asset_type,
                file,
                credential_ids,
                runner,
                sboms,
            )

            ci_logger.output(name="scan_id", value=scan_id)
            ci_logger.info(f"Scan created with id {scan_id}.")

            if break_on_risk_rating is not None and break_on_risk_rating != "":
                run.apply_break_scan_risk_rating(
                    break_on_risk_rating, scan_id, max_wait_minutes, runner, ci_logger
                )

        except base_runner.ResponseError as e:
            ci_logger.error(f"Could not start the scan. {e}")
            raise click.exceptions.Exit(2) from e
        except TimeoutError:
            ci_logger.error("The scan is still running.")
            raise click.exceptions.Exit(2) from None
    else:
        ci_logger.error("API key not not provided.")
        raise click.exceptions.Exit(2) from None


def _create_scan(title, scan_profile, asset_type, file, credential_ids, runner, sboms):
    scan_result = runner.execute(
        scan_create_api.CreateMobileScanAPIRequest(
            title=title,
            asset_type=asset_type,
            scan_profile=scan_profile,
            application=file,
            test_credential_ids=credential_ids,
            sboms=sboms,
        )
    )
    scan_id = scan_result.get("data").get("createMobileScan").get("scan").get("id")
    return scan_id


def _create_test_credentials(test_credentials, runner):
    credential_ids = []
    for test_credential in test_credentials:
        test_credential_result = runner.execute(
            test_credentials_create_api.CreateTestCredentialAPIRequest(
                test_credential=test_credential
            )
        )
        credential_ids.append(
            test_credential_result.get("data")
            .get("createTestCredentials")
            .get("testCredentials")
            .get("id")
        )
    return credential_ids



"""Vulnz Describe command."""
import logging
from typing import Optional

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.vulnz import vulnz
from ostorlab.runtimes.cloud import runtime as cloud_runtime
from ostorlab.runtimes.local import runtime as local_runtime

console = cli_console.Console()

logger = logging.getLogger(__name__)


@vulnz.command(name="describe")
@click.option(
    "--vuln-id", "-v", "vuln_id", help="Id of the vulnerability.", required=False
)
@click.option("--scan-id", "-s", "scan_id", help="Id of the scan.", required=False)
@click.pass_context
def describe_cli(
    ctx, vuln_id: Optional[int] = None, scan_id: Optional[int] = None
) -> None:
    """Describe a vuln by id (--vuln-id) or all the vulnz for as specific scan.
    example of usage:
        - ostorlab vulnz --runtime cloud describe --scan-id 54821"""
    if isinstance(ctx.obj["runtime"], cloud_runtime.CloudRuntime) and scan_id is None:
        raise click.BadParameter(
            "Please provide --scan-id / -s when using the cloud runtime"
        )
    if (
        isinstance(ctx.obj["runtime"], local_runtime.LocalRuntime)
        and scan_id is None
        and vuln_id is None
    ):
        raise click.BadParameter(
            "Please provide --scan-id / -s or --vuln-id / -v when using local runtime"
        )
    runtime_instance = ctx.obj["runtime"]
    console.info("Fetching vulnerabilities...")
    runtime_instance.describe_vuln(scan_id=scan_id, vuln_id=vuln_id)



"""Asset of type Link."""
import logging
from typing import List

import click

from ostorlab.assets import link as link_asset
from ostorlab.cli.scan.run import run
from ostorlab.cli import console as cli_console


logger = logging.getLogger(__name__)
console = cli_console.Console()


@run.run.command()
@click.option("--url", help="List of Urls to scan.", required=True, multiple=True)
@click.option("--method", help="List of HTTP methods.", required=True, multiple=True)
@click.pass_context
def link(ctx: click.core.Context, url: List[str], method: List[str]) -> None:
    """Run scan for links."""
    runtime = ctx.obj["runtime"]
    if len(url) != len(method):
        console.error("Make sure every URL has its corresponding method.")
        raise click.exceptions.Exit(2)
    assets = []
    for u, m in zip(url, method):
        asset = link_asset.Link(url=u, method=m)
        assets.append(asset)
    logger.debug("scanning assets %s", asset)
    runtime.scan(
        title=ctx.obj["title"],
        agent_group_definition=ctx.obj["agent_group_definition"],
        assets=assets,
    )



"""Logger is in charge of defining the output format based on CI expected output."""
import abc


class Logger(abc.ABC):
    """Logger is in charge of printing the results based on the CI expected output."""

    @abc.abstractmethod
    def info(self, message: str) -> None:
        """Print Info messages.

        Args:
            message: message to print.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def error(self, message: str) -> None:
        """Print Error messages.

        Args:
            message: message to print.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def output(self, name: str, value: str) -> None:
        """Pass an output to the next step of the CI.

        Args:
            name: name of the output to pass to the next step.
            value: value of the output.
        """
        raise NotImplementedError()



"""Agent class.

All agents should inherit from the agent class to access the different features, like automated message
serialization, message receiving and sending, selector enrollment, agent health check, etc.

To use it, check out documentations at https://docs.ostorlab.co/.
"""
import abc
import argparse
import asyncio
import base64
import json
import logging
import os
import pathlib
import signal
import sys
import threading
import uuid
from typing import Dict, Any, Optional, Type, List

import google.cloud.logging
from google.oauth2 import service_account

from ostorlab import exceptions
from ostorlab.agent import definitions as agent_definitions
from ostorlab.agent.message import message as agent_message
from ostorlab.agent.mixins import agent_healthcheck_mixin
from ostorlab.agent.mixins import agent_mq_mixin
from ostorlab.agent.mixins import agent_open_telemetry_mixin as open_telemetry_mixin
from ostorlab.runtimes import definitions as runtime_definitions
from ostorlab.utils import system

GCP_LOGGING_CREDENTIAL_ENV = "GCP_LOGGING_CREDENTIAL"

AGENT_DEFINITION_PATH = "/tmp/ostorlab.yaml"

logger = logging.getLogger(__name__)


class NonListedMessageSelectorError(exceptions.OstorlabError):
    """Emit selector is not listed in the out_selector list."""


class MaximumCyclicProcessReachedError(exceptions.OstorlabError):
    """The cyclic process limit is enforced and reach set value."""


class MaximumDepthProcessReachedError(exceptions.OstorlabError):
    """The processing depth limit is enforced and reached the limit."""


def _setup_logging(agent_key: str, universe: str) -> None:
    gcp_logging_credential = os.environ.get(GCP_LOGGING_CREDENTIAL_ENV)
    if gcp_logging_credential is not None:
        info = json.loads(base64.b64decode(gcp_logging_credential.encode()).decode())
        credentials = service_account.Credentials.from_service_account_info(info)
        client = google.cloud.logging.Client(credentials=credentials)
        client.setup_logging(labels={"agent_key": agent_key, "universe": universe})


class AgentMixin(
    agent_mq_mixin.AgentMQMixin, agent_healthcheck_mixin.AgentHealthcheckMixin, abc.ABC
):
    """Agent mixin handles all the heavy lifting.

    The agent mixin start the healthcheck service, connects the MQ and start listening to the process message.
    """

    def __init__(
        self,
        agent_definition: agent_definitions.AgentDefinition,
        agent_settings: runtime_definitions.AgentSettings,
    ) -> None:
        """Inits the agent configuration from the Yaml agent definition.

        Args:
            agent_definition: Agent definition dictating the settings of the agent, like name, in_selectors ...
            agent_settings: The running instance definition dictating custom settings of the agent like bus
             URL.
        """
        self._loop = asyncio.get_event_loop()
        self._agent_definition = agent_definition
        self._agent_settings = agent_settings
        self._control_message: Optional[agent_message.Message] = None
        self.name = agent_definition.name
        self.in_selectors = agent_definition.in_selectors
        self.out_selectors = agent_definition.out_selectors
        self.cyclic_processing_limit = agent_settings.cyclic_processing_limit
        self.depth_processing_limit = agent_settings.depth_processing_limit
        # Arguments are defined in the agent definition, and can have a default value. The value can also be set from
        # the scan definition in the agent group. Therefore, we read both and override the value from the passed args.
        self.defined_args = agent_definition.args
        self.passed_args = agent_settings.args
        self.bus_url = agent_settings.bus_url
        self.bus_exchange_topic = agent_settings.bus_exchange_topic
        self.bus_managment_url = agent_settings.bus_management_url
        self.bus_vhost = agent_settings.bus_vhost
        agent_mq_mixin.AgentMQMixin.__init__(
            self,
            name=agent_definition.name,
            # Selectors are mapped to queue binding that listen to all
            # sub-routing keys.
            keys=[f"{s}.#" for s in self.in_selectors],
            url=self.bus_url,
            topic=self.bus_exchange_topic,
        )
        agent_healthcheck_mixin.AgentHealthcheckMixin.__init__(
            self,
            name=agent_definition.name,
            host=agent_settings.healthcheck_host,
            port=agent_settings.healthcheck_port,
        )
        signal.signal(signal.SIGTERM, self._handle_signal)

    @property
    def definition(self) -> agent_definitions.AgentDefinition:
        """Agent definition property."""
        return self._agent_definition

    @property
    def settings(self) -> runtime_definitions.AgentSettings:
        """Agent settings property."""
        return self._agent_settings

    @property
    def args(self) -> Dict[str, Any]:
        """Agent arguments as passed from definition and settings."""
        arguments = {}
        # First read the agent default values.
        for a in self.definition.args:
            arguments[a["name"]] = a.get("value")
        # Override the default values from settings.
        for a in self.settings.args:
            # Enforce that only declared arguments are accepted.
            if a.name not in arguments:
                # TODO(OS-5119): Change behavior to fail of the argument is missing from definition.
                logger.warning(
                    "Argument %s is defined in the agent settings but not in the agent definition. "
                    "Please update your definition file or the agent will fail in the future.",
                    a.name,
                )

            if a.type == "binary":
                arguments[a.name] = a.value
            else:
                arguments[a.name] = json.loads(a.value.decode())

        return arguments

    @property
    def universe(self) -> Optional[str]:
        """Returns the current scan universe.

        A universe is the group of agents and services in charge of running a scan. The universe is defined
        by the runtime."""
        return os.environ.get("UNIVERSE")

    def run(self) -> None:
        """Starts running the agent.

        Connects to the agent bus, start health check and start listening to new messages.
        """
        self.add_healthcheck(self.is_healthy)
        self.start_healthcheck()
        self._loop.run_until_complete(self.mq_init())
        logger.debug("calling start method")
        # This is call in a thread to avoid blocking calls from affecting the MQ heartbeat running on the main thread.
        t = threading.Thread(target=self.start)
        t.start()
        t.join()
        logger.debug("calling start method done")
        try:
            if self.in_selectors is not None and len(self.in_selectors) > 0:
                logger.debug("starting mq run")
                self._loop.run_until_complete(self.mq_run())
                self._loop.run_forever()
        finally:
            logger.debug("closing bus and loop")
            self._loop.close()

    @abc.abstractmethod
    def is_healthy(self) -> bool:
        """Overridable agent health check method to add custom health check logic.

        Returns:
            bool to indicate if the agent is healthy of not.
        """
        raise NotImplementedError()

    def process_message(self, selector: str, message: bytes) -> None:
        """Processes raw message received from BS.

        Args:
            selector: destination selector with full path, including UUID set by default.
            message: raw bytes message.

        Returns:
            None
        """

        self._control_message = agent_message.Message.from_raw("v3.control", message)
        raw_message = self._control_message.data["message"]
        # remove the UUID from the selector:
        selector = ".".join(selector.split(".")[:-1])
        object_message = agent_message.Message.from_raw(selector, raw_message)

        # Validate the message before processing it.
        try:
            self._validate_message()
        except MaximumCyclicProcessReachedError:
            self.on_max_cyclic_process_reached(object_message)
        except MaximumDepthProcessReachedError:
            self.on_max_depth_process_reached(object_message)

        try:
            logger.debug("Call to process with message= %s", raw_message)
            self.process(object_message)
        except Exception as e:  # pylint: disable="broad-except"
            system_info = system.get_system_info()
            if system_info is not None:
                logger.error("System Info: %s", system_info)
            logger.error("Message: %s", object_message)
            logger.exception("Exception: %s", e)
        finally:
            self.process_cleanup()
            logger.debug("done call to process message")
            # Flush all logging handlers to ensure remote logging is sent before app shutdown.
            for h in logger.handlers:
                h.flush()

    def _validate_message(self) -> None:
        """Check the message received is valid, currently only check for cyclic processing limit."""
        control_agents: list[str] = self._control_message.data["control"]["agents"]
        if (
            self.cyclic_processing_limit is not None
            and self.cyclic_processing_limit != 0
        ):
            if control_agents.count(self.name) >= self.cyclic_processing_limit:
                raise MaximumCyclicProcessReachedError()

        if self.depth_processing_limit is not None and self.depth_processing_limit != 0:
            if len(control_agents) >= self.depth_processing_limit:
                agent_path = " -> ".join(control_agents)
                error_message = (
                    f"The maximum depth processing limit of {self.depth_processing_limit} agents is reached. "
                    f"Agents path: {agent_path}"
                )
                raise MaximumDepthProcessReachedError(error_message)

    @abc.abstractmethod
    def process_cleanup(self) -> None:
        """Overridable message cleanup method to be called once process is completed or even in the case of a failure.

        Returns:
            None
        """
        raise NotImplementedError()

    def _handle_signal(self, *args) -> None:
        """Call the Agent `at_exit` method responsible for executing cleanup
        in the case of unexpected agent termination.
        """
        del args
        self.at_exit()
        sys.exit()

    @abc.abstractmethod
    def at_exit(self) -> None:
        """Overridable at exit method to perform cleanup in the case of expected and unexpected agent termination.

        Returns:

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def start(self) -> None:
        """Overridable agent start to implement one-off or long-processing non-receiving agents.

        Returns:
            None
        """
        raise NotImplementedError("Missing process method implementation.")

    @abc.abstractmethod
    def process(self, message: agent_message.Message) -> None:
        """Overridable message processing method.

        Args:
            message: message received from with selector and data.

        Returns:
            None
        """
        raise NotImplementedError("Missing process method implementation.")

    def emit(
        self, selector: str, data: Dict[str, Any], message_id: Optional[str] = None
    ) -> None:
        """Sends a message to all listening agents on the specified selector.

        Args:
            selector: target selector.
            data: message data to be serialized.
            message_id: An id that will be added to the tail of the message.
        Raises:
            NonListedMessageSelectorError: when selector is not part of listed out selectors.

        Returns:
            None
        """
        message = agent_message.Message.from_data(selector, data)
        self.emit_raw(selector, message.raw, message_id=message_id)

    @abc.abstractmethod
    def on_max_cyclic_process_reached(self, message: agent_message.Message) -> None:
        """Overridable method triggered on max cyclic process reached.

        Returns:
            None
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def on_max_depth_process_reached(self, message: agent_message.Message) -> None:
        """Overridable method triggered on max processing depth reached."""
        raise NotImplementedError()

    def emit_raw(
        self, selector: str, raw: bytes, message_id: Optional[str] = None
    ) -> None:
        """Sends a message to all listening agents on the specified selector with no serialization.

        Args:
            selector: target selector.
            raw: raw message to send.
            message_id: An id that will be added to the tail of the message.
        Raises:
            NonListedMessageSelectorError: when selector is not part of listed out selectors.

        Returns:
            None
        """
        if selector not in self.out_selectors:
            logger.error("selector not present in list of out selectors")
            # CAUTION: this check is enforced on the client-side only in certain runtimes
            raise NonListedMessageSelectorError(
                f'{selector} is not in {"".join(self.out_selectors)}'
            )

        logger.debug("call to send message with %s", selector)
        # A random unique UUID is added to ensure messages could be resent. Storage master ensures that a message with
        # the same selector and message body is sent only once to the bus.
        if message_id is None:
            selector = f"{selector}.{uuid.uuid4()}"
        else:
            selector = f"{selector}.{message_id}"

        control_message = self._prepare_message(raw)
        self.mq_send_message(selector, control_message)
        logger.debug("done call to send_message")

    def _prepare_message(self, raw: bytes) -> bytes:
        if self._control_message is not None:
            agents = [*self._control_message.data["control"]["agents"], self.name]
        else:
            agents = [self.name]
        control_message = agent_message.Message.from_data(
            "v3.control", {"control": {"agents": agents}, "message": raw}
        )
        return control_message.raw

    @classmethod
    def main(cls: Type["AgentMixin"], args: Optional[List[str]] = None) -> None:
        """Prepares the agents class by reading the agent definition and runtime settings.

        By the default, the class main expects the definition file to be at `agent.yaml` and settings to be at
        `/tmp/settings.binproto`.

        The values can be overridden by passing the arguments `--definition` and `--settings`.

        The definition file defines the agents, like name, description, consumed and generated selectors. The definition
        might include information on how to run the agent, but those can be overridden by the scan runtime.

        The settings file defines how the agent is running, what services are enabled and what addresses should it
        connect to. Some of these settings are consumed by the scan runtime, others are consumed by the agent itself.

        Remote Logging:
            If "GCP_LOGGING_CREDENTIAL" is present in the env variables, the Agent will use it to enable
            remote logging to a GCP project.

        Args:
            args: Arguments passed to the argument parser. These are added for testability.

        Returns:
            The function do not return unless an error is detected. The agent instance starts running and never returns.
        """

        # Settings the args to make the code testable without mocking.
        if not args:
            args = sys.argv[1:]

        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-s",
            "--settings",
            default="/tmp/settings.binproto",
            help="Agent binary proto settings.",
        )
        parsed_args = parser.parse_args(args)
        logger.info(
            "running agent with definition %s and settings %s",
            AGENT_DEFINITION_PATH,
            parsed_args.settings,
        )

        if not pathlib.Path(AGENT_DEFINITION_PATH).exists():
            logger.error("definition file does not exist")
            sys.exit(2)
        if not pathlib.Path(parsed_args.settings).exists():
            logger.error("settings file does not exist")
            sys.exit(2)

        with open(AGENT_DEFINITION_PATH, "r", encoding="utf-8") as f_definition, open(
            parsed_args.settings, "rb"
        ) as f_settings:
            agent_definition = agent_definitions.AgentDefinition.from_yaml(f_definition)
            agent_settings = runtime_definitions.AgentSettings.from_proto(
                f_settings.read()
            )

            _setup_logging(
                agent_key=agent_settings.key, universe=os.environ.get("UNIVERSE")
            )

            instance = cls(
                agent_definition=agent_definition, agent_settings=agent_settings
            )
            logger.debug("running agent instance")
            instance.run()


class Agent(open_telemetry_mixin.OpenTelemetryMixin, AgentMixin):
    """Agent class.

    An agent can either be a message processor or standalone. Standalone agents can either be long-running or run-once.

    Message processor must implement the process method and declare a set of listen selectors. The selectors are defined
    in the YAML agent definition file.

    Standalone agents don't have any selectors to listen to and must implement its logic in the start method. Use-cases
    include input injectors, like asset injector or configuration injectors. It also includes long-running processes,
    like a proxy.

    Agents can optionally define a custom health check method `is_healthy` to inform the run time that the agent is
    operational and ensure the scan asset is not injected before the agent has completed his setup.

    The agent can also define a `process_cleanup` method that runs after every process method, even in the case of an
    exception. This can be used to perform actions, like flushing messages or status recovery.

    `at_exit` can also be defined to execute before the agent exits. This method is not 100% guaranteed before the agent
    stops. The function is not called when the program is killed by a signal not handled by Python, when a Python fatal
    internal error is detected, or when `os._exit()` is called.

    For example and more details on how to implement an agent, refer to https://docs.ostorlab.co.
    """

    def is_healthy(self) -> bool:
        """Overridable agent health check method to add custom health check logic.

        Returns:
            bool to indicate if the agent is healthy or not.
        """
        return True

    def start(self) -> None:
        """Overridable agent start to implement one-off or long-processing non-receiving agents.

        Returns:
            None
        """
        pass

    def process(self, message: agent_message.Message) -> None:
        """Overridable message processing method.

        Args:
            message: message received from with selector and data.

        Returns:
            None
        """
        raise NotImplementedError("Missing process method implementation.")

    def process_cleanup(self) -> None:
        """Overridable message cleanup method to be called once process is completed or even in the case of a failure.

        Returns:
            None
        """
        pass

    def at_exit(self) -> None:
        """Overridable at exit method to perform cleanup in the case of expected and unexpected agent termination.

        Returns:
            None
        """
        pass

    def on_max_cyclic_process_reached(self, message: agent_message.Message) -> None:
        """Overridable method triggered on max cyclic process reached.

        Returns:
            None
        """
        pass

    def on_max_depth_process_reached(self, message: agent_message.Message) -> None:
        """Overridable method triggered on max processing depth reached."""
        pass



"""Agent List command."""
import logging

import docker

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent

console = cli_console.Console()

logger = logging.getLogger(__name__)


def _human_readable_size(size, decimal_places=3):
    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if size < 1024.0:
            break
        size /= 1024.0
    return f"{size:.{decimal_places}f}{unit}"


@agent.command(name="list")
def list_cli() -> None:
    """CLI command to list installed agents."""
    docker_client = docker.from_env()
    images = docker_client.images.list()
    console.success("Agents listed successfully.")
    agents = []
    for im in images:
        for t in im.tags:
            if t.startswith("agent_"):
                agents.append(
                    {
                        "id": im.short_id,
                        "agent": t.replace("_", "/", 2).split(":")[0],
                        "version": t.split(":")[1],
                        "size": _human_readable_size(im.attrs["Size"]),
                        "created": im.attrs["Created"],
                    }
                )

    columns = {
        "Agent": "agent",
        "Version": "version",
        "Id": "id",
        "Size": "size",
        "Created": "created",
    }
    title = f"Found {len(agents)} Agents"
    console.table(columns=columns, data=agents, title=title)



"""Helper functions to get system information."""
from dataclasses import dataclass
from typing import Optional


@dataclass
class CpuInfo:
    """CPU information."""

    load: list[float]


@dataclass
class MemoryInfo:
    """Memory information."""

    total: int
    available: int
    used: int
    percent: float


@dataclass
class StorageInfo:
    """Storage information."""

    total: int
    used: int
    free: int


@dataclass
class SystemLoadInfo:
    """System load information."""

    cpu_load: CpuInfo
    memory_usage: MemoryInfo
    storage: StorageInfo

    def __str__(self) -> str:
        """String representation of the system load information."""
        return (
            f"CPU Load: {self.cpu_load.load}\n"
            f"Memory Usage: Total={self.memory_usage.total}, "
            f"Available={self.memory_usage.available}, "
            f"Used={self.memory_usage.used}, "
            f"Percent={self.memory_usage.percent}\n"
            f"Storage: Total={self.storage.total}, "
            f"Used={self.storage.used}, "
            f"Free={self.storage.free}"
        )


def get_system_info() -> Optional[SystemLoadInfo]:
    """Get system load information.

    We use psutil to get the system load information. in case `psutil` is not available, we return None.

    Returns:
        System load information or None in case `psutil` is not available.
    """
    try:
        # Import psutil here to catch the ImportError in case psutil is not available.
        # pylint: disable=import-outside-toplevel
        import psutil

        # Get CPU load information
        cpu_info = CpuInfo(psutil.cpu_percent(interval=None, percpu=True))

        # Get memory usage information
        virtual_memory = psutil.virtual_memory()
        memory_info = MemoryInfo(
            total=virtual_memory.total,
            available=virtual_memory.available,
            used=virtual_memory.used,
            percent=virtual_memory.percent,
        )

        # Get storage information
        partitions = psutil.disk_partitions()
        total_storage, used_storage, free_storage = 0, 0, 0
        for partition in partitions:
            partition_usage = psutil.disk_usage(partition.mountpoint)
            total_storage += partition_usage.total
            used_storage += partition_usage.used
            free_storage += partition_usage.free
        storage_info = StorageInfo(
            total=total_storage, used=used_storage, free=free_storage
        )

        # Create the SystemLoadInfo data class instance
        return SystemLoadInfo(
            cpu_load=cpu_info, memory_usage=memory_info, storage=storage_info
        )

    except ImportError:
        # psutil is not available, return None
        return None



"""Auth login command."""

import logging

import click

from ostorlab.apis.runners import authenticated_runner
from ostorlab.cli.auth import auth
from ostorlab.cli import console as cli_console

logger = logging.getLogger(__name__)
console = cli_console.Console()


@auth.auth.command()
@click.option(
    "--username", "-u", help="Ostorlab platform username.", required=True, prompt=True
)
@click.option(
    "--password",
    "-p",
    help="Ostorlab platform password.",
    required=True,
    prompt=True,
    hide_input=True,
)
@click.option(
    "--token-duration",
    help="Expiration time for token (m for minutes, h for hours, and d for days).",
)
def login(username, password, token_duration):
    """Use this to log into your account."""
    try:
        api_runner = authenticated_runner.AuthenticatedAPIRunner(
            username=username, password=password, token_duration=token_duration
        )
        api_runner.authenticate()
    except authenticated_runner.AuthenticationError:
        console.error(
            "Authentication error, please check that your credentials are valid."
        )



"""Module responsible for dumping data to different formats."""

import abc
import csv
import json
from typing import List

FIELDNAMES = [
    "id",
    "title",
    "location",
    "risk_rating",
    "cvss_v3_vector",
    "short_description",
    "description",
    "recommendation",
    "references",
    "technical_detail",
]


class VulnzDumper(abc.ABC):
    """Dumper Base class: All dumpers should inherit from this class to access the dump method."""

    def __init__(self, output_path: str) -> None:
        """Constructs all the necessary attributes for the object.

        Args:
            output_path: path to the output file.
            data: dictionary, with vulnerability id as a key,
                  and a dictionary of the vulnerability information as values.

        Returns:
            None
        """
        self.output_path: str = output_path

    @abc.abstractmethod
    def dump(self, data: List[object]) -> None:
        """Dump the vulnerabilities in the right format."""
        raise NotImplementedError("Missing implementation")


class VulnzJsonDumper(VulnzDumper):
    """Vulnerability dumper to json."""

    def dump(self, data: List[object]) -> None:
        """Dump vulnerabilities to json file.

        Raises:
            FileNotFoundError: in case the path or file name are invalid.
        """
        if not self.output_path.endswith(".jsonl"):
            self.output_path += ".jsonl"
        with open(self.output_path, "a", encoding="utf-8") as f:
            for item in data:
                f.write(json.dumps(item) + "\n")


class VulnzCsvDumper(VulnzDumper):
    """Vulnerability dumper to csv."""

    def dump(self, data: List[object]) -> None:
        """Dump vulnerabilities to csv file.

        Raises:
            FileNotFoundError: in case the path or file name are invalid.
        """
        if not self.output_path.endswith(".csv"):
            self.output_path += ".csv"
        with open(self.output_path, "a", encoding="utf-8") as outfile:
            csv_writer = csv.DictWriter(outfile, fieldnames=FIELDNAMES)
            csv_writer.writeheader()
            for item in data:
                csv_writer.writerow(item)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/domain_name/service/service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ostorlab/agent/message/proto/v3/asset/domain_name/service/service.proto',
  package='ostorlab.agent.message.proto.v3.asset.domain_name.service',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\nGostorlab/agent/message/proto/v3/asset/domain_name/service/service.proto\x12\x39ostorlab.agent.message.proto.v3.asset.domain_name.service\"D\n\x07Message\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x0e\n\x06schema\x18\x03 \x01(\t\x12\r\n\x05state\x18\x04 \x01(\t'
)




_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='ostorlab.agent.message.proto.v3.asset.domain_name.service.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='ostorlab.agent.message.proto.v3.asset.domain_name.service.Message.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='ostorlab.agent.message.proto.v3.asset.domain_name.service.Message.port', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='ostorlab.agent.message.proto.v3.asset.domain_name.service.Message.schema', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='ostorlab.agent.message.proto.v3.asset.domain_name.service.Message.state', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=134,
  serialized_end=202,
)

DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGE,
  '__module__' : 'ostorlab.agent.message.proto.v3.asset.domain_name.service.service_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.asset.domain_name.service.Message)
  })
_sym_db.RegisterMessage(Message)


# @@protoc_insertion_point(module_scope)
